import re
from typing import Dict, List, Optional, Union

from ufal.udpipe import InputFormat
from ufal.udpipe import Model
from ufal.udpipe import OutputFormat, ProcessingError, Sentence, Word

from .utils import get_path

NO_SPACE = "SpaceAfter=No"


class PretokenizedInputFormat(object):
    """Dummy tokenizer for pretokenized input.

    Execution speed might be slow compared to other UDPipe tokenizers
    due to pure Python implementation. Mocks InputFormat API to enable
    plug-and-play behaviour.
    """

    def setText(self, text: str) -> None:
        """Store text in iterable lines for tokenization.

        text: string, where each sentence is on a line and tokens
              are separated by tabs.
        """
        self.lines = iter(text.split("\n"))

    def nextSentence(self, sentence: Sentence, _: ProcessingError) -> bool:
        """Tokenize each line from stored lines and store tokens in sentence.

        sentence: UDPipe container for storing tokens.
        """
        try:
            line = next(self.lines)
        except StopIteration:
            return False
        tokens = line.split("\t")
        prev_word = Word()
        for token in tokens:
            word = sentence.addWord(token)
            if re.match(r"\W", token):
                # leave no space after previous token iff current token
                # is non-alphanumeric (i.e. punctuation)
                prev_word.misc = NO_SPACE
            prev_word = word
        return True


class UDPipeModel(object):

    def __init__(
        self,
        lang: str,
        path: Optional[str] = None,
        meta: Optional[Dict] = None
    ):
        """Load UDPipe model for given language.

        lang: ISO 639-1 language code or shorthand UDPipe model name.
        path: Path to UDPipe model.
        meta: Meta-information about the UDPipe model.
        """
        path = path or get_path(lang=lang)
        self.model = Model.load(path)
        self._lang = lang.split("-")[0]
        self._meta = meta or {"author": "Milan Straka & Jana Straková",
                              "description": "UDPipe pretrained model.",
                              "email": "straka@ufal.mff.cuni.cz",
                              "lang": f"udpipe_{self._lang}",
                              "license": "CC BY-NC-SA 4.0",
                              "name": path.split("/")[-1],
                              "parent_package": "spacy_udpipe",
                              "pipeline": [
                                  "Tokenizer", "Tagger", "Lemmatizer", "Parser"
                              ],
                              "source": "Universal Dependencies 2.5",
                              "url": "http://ufal.mff.cuni.cz/udpipe",
                              "version": "1.2.0"
                              }

    def __call__(
        self,
        text: Union[
            str,
            List[str],
            List[List[str]]
        ]
    ) -> List[Sentence]:
        """Tokenize, tag and parse the text and return it in an UDPipe
        representation.

        text: Input text, can be presegmented or pretokenized:
            str             : raw text,
            List[str]       : presegmented text,
            List[List[str]] : pretokenized text.
        RETURNS: Processed sentences.
        """
        sentences = self.tokenize(text)
        for s in sentences:
            self.tag(s)
            self.parse(s)
        return sentences

    def _read(self, text: str, input_format: str) -> List[Sentence]:
        """Convert the text to an UDPipe representation.

        text: Input text.
        input_format: Desired input format.
        RETURNS: Processed sentences.
        """
        input_format.setText(text)
        error = ProcessingError()
        sentences = []

        sentence = Sentence()
        while input_format.nextSentence(sentence, error):
            sentences.append(sentence)
            sentence = Sentence()
        if error.occurred():
            raise Exception(error.message)

        return sentences

    def tokenize(
        self,
        text: Union[
            str,
            List[str],
            List[List[str]]
        ]
    ) -> List[Sentence]:
        """Tokenize input text.

        text: Input text, can be presegmented or pretokenized:
            str             : raw text,
            List[str]       : presegmented text,
            List[List[str]] : pretokenized text.
        Note: both presegmented and pretokenized text can not contain
              newline or tab characters.
        RETURNS: Processed sentences.
        """
        if isinstance(text, str):
            tokenizer = self.model.newTokenizer(self.model.DEFAULT)
        elif isinstance(text, list):
            if isinstance(text[0], list):
                text = "\n".join("\t".join(sent) for sent in text)
                tokenizer = PretokenizedInputFormat()
            else:
                text = "\n".join(text)
                tokenizer = self.model.newTokenizer(
                    self.model.TOKENIZER_PRESEGMENTED
                )
        else:
            raise TypeError(
                "\n".join(
                    (f"Input type is {type(text)}, but must be one:",
                     "str             : raw text",
                     "List[str]       : presegmented text",
                     "List[List[str]] : pretokenized text")
                )
            )
        if not tokenizer:
            raise Exception(
                "The model does not have a tokenizer "
                f"so it can not tokenize input: {text}"
            )
        return self._read(text=text, input_format=tokenizer)

    def tag(self, sentence: Sentence) -> None:
        """Assign part-of-speech tags (inplace).

        sentence: Input sentence.
        """
        self.model.tag(sentence, self.model.DEFAULT)

    def parse(self, sentence: Sentence) -> None:
        """Assign dependency parse relations (inplace).

        sentence: Input sentence.
        """
        self.model.parse(sentence, self.model.DEFAULT)

    def read(self, text: str, in_format: str) -> List[Sentence]:
        """Load text in the given format and return it in an UDPipe
        representation.

        text: Text to load.
        in_format: 'conllu'|'horizontal'|'vertical'.
        RETURNS: Processed sentences.
        """
        input_format = InputFormat.newInputFormat(in_format)
        if not input_format:
            raise Exception(f"Cannot create input format '{in_format}'")
        return self._read(text=text, input_format=input_format)

    def write(self, sentences: List[Sentence], out_format: str) -> str:
        """Write given sentences in the required output format.

        sentences: Input ufal.udpipe.Sentence-s.
        out_format: 'conllu'|'horizontal'|'vertical'.
        RETURNS: Sentences formatted in the out_format.
        """
        output_format = OutputFormat.newOutputFormat(out_format)
        output = "".join([output_format.writeSentence(s) for s in sentences])
        output += output_format.finishDocument()

        return output



import json
import os
import urllib.request
from typing import Optional, Dict

from spacy import blank, Language
from spacy.util import get_lang_class


BASE_URL = "https://lindat.mff.cuni.cz/repository/xmlui/bitstream/handle/11234/1-3131"  # noqa: E501
MODELS_DIR = os.path.join(os.path.dirname(__file__), "models")
with open(
    os.path.join(
        os.path.dirname(__file__),
        "languages.json"
    ),
) as f:
    LANGUAGES = json.load(f)


def _check_language(lang: str) -> None:
    assert lang in LANGUAGES, f"'{lang}' language not available"


def _check_models_dir(lang: str) -> None:
    assert os.path.exists(MODELS_DIR), "Download the pretrained model(s) first"


def download(lang: str) -> None:
    """Download the UDPipe pretrained model.

    lang: ISO 639-1 language code or shorthand UDPipe model name.
    """
    _check_language(lang)
    try:
        _check_models_dir(lang)
    except AssertionError:
        os.makedirs(MODELS_DIR)
    if LANGUAGES[lang] in os.listdir(MODELS_DIR):
        print(f"Already downloaded a model for the '{lang}' language")
        return
    url = f"{BASE_URL}/{LANGUAGES[lang]}"
    filename = os.path.join(MODELS_DIR, LANGUAGES[lang])
    urllib.request.urlretrieve(url=url, filename=filename)
    print(f"Downloaded pre-trained UDPipe model for '{lang}' language")


def get_path(lang: str) -> str:
    """Get the path to the UDPipe pretrained model if it was downloaded.

    lang: ISO 639-1 language code or shorthand UDPipe model name.
    RETURNS: The path to the UDPipe pretrained model.
    """
    _check_language(lang)
    _check_models_dir(lang)
    if not LANGUAGES[lang] in os.listdir(MODELS_DIR):
        raise Exception(
                "Use spacy_udpipe.download to download the pre-trained"
                f" UDPipe model for the '{lang}' language"
            )
    path = os.path.join(MODELS_DIR, LANGUAGES[lang])
    return path


def get_defaults(lang: str) -> Language.Defaults:
    """Get the language-specific defaults, if available in spaCy. This allows
    using lexical attribute getters that depend on static language data, e.g.
    Token.like_num, Token.is_stop, Doc.noun_chunks, etc.

    lang: ISO 639-1 language code or shorthand UDPipe model name.
    RETURNS: The language defaults.
    """
    try:
        lang_cls = get_lang_class(lang)
        return lang_cls.Defaults
    except ImportError:
        return Language.Defaults


def load(
    lang: str = ""
) -> Language:
    """Convenience function for initializing the Language class that
    mimicks spacy.load.

    lang: ISO 639-1 language code or shorthand UDPipe model name.
    RETURNS: SpaCy Language object with UDPipeTokenizer.
    """
    config = {"nlp": {"tokenizer": {}}}
    name = lang.split("-")[0]
    config["nlp"]["tokenizer"]["@tokenizers"] = "spacy_udpipe.PipelineAsTokenizer.v1"  # noqa: E501
    # Set UDPipe options
    config["nlp"]["tokenizer"]["lang"] = lang
    config["nlp"]["tokenizer"]["path"] = get_path(lang)
    config["nlp"]["tokenizer"]["meta"] = None
    return blank(name, config=config)


def load_from_path(
    lang: str,
    path: str,
    meta: Optional[Dict] = {"description": "custom model"},
) -> Language:
    """Convenience function for initializing the Language class
    and loading a custom UDPipe model via the path argument.

    lang: ISO 639-1 language code or shorthand UDPipe model name.
    path: Path to the UDPipe model.
    meta: Optional meta-information about the UDPipe model.
    RETURNS: SpaCy Language object with UDPipeTokenizer.
    """
    config = {"nlp": {"tokenizer": {}}}
    name = lang.split("-")[0]
    config["nlp"]["tokenizer"]["@tokenizers"] = "spacy_udpipe.PipelineAsTokenizer.v1"  # noqa: E501
    # Set UDPipe options
    config["nlp"]["tokenizer"]["lang"] = lang
    config["nlp"]["tokenizer"]["path"] = path
    config["nlp"]["tokenizer"]["meta"] = meta
    return blank(name, config=config)



__version__ = "1.0.0"
__all__ = [
    "download", "load", "load_from_path",
    "UDPipeTokenizer", "UDPipeModel"
]

from .utils import download, load, load_from_path
from .tokenizer import UDPipeTokenizer
from .udpipe import UDPipeModel



import re
from typing import Optional, Iterable, List, Tuple, Union, Dict

from spacy.tokens import Doc
from spacy.vocab import Vocab
from spacy.util import registry
from ufal.udpipe import Sentence, Word

from .udpipe import NO_SPACE, UDPipeModel
from .utils import get_path


@registry.tokenizers("spacy_udpipe.PipelineAsTokenizer.v1")
def create_tokenizer(
    lang: str = "",
    path: Optional[str] = None,
    meta: Optional[Dict] = None
):
    def tokenizer_factory(
        nlp,
        lang=lang,
        path=path,
        meta=meta
    ) -> UDPipeTokenizer:
        model = UDPipeModel(
            lang=lang,
            path=path or get_path(lang),
            meta=meta
        )
        return UDPipeTokenizer(
            model=model,
            vocab=nlp.vocab
        )

    return tokenizer_factory


def _spacy_dep(d: str) -> str:
    # Ensure labels match with SpaCy
    return d.upper() if d == "root" else d


class UDPipeTokenizer(object):
    """Custom Tokenizer which sets all the attributes because
    the UDPipe pipeline runs only once and does not
    contain separate spaCy pipeline components.
    """

    _ws_pattern = re.compile(r"\s+")

    def __init__(
        self,
        model: UDPipeModel,
        vocab: Vocab
    ):
        """Initialize the tokenizer.

        model: The initialized UDPipe model.
        vocab: The vocabulary to use.
        """
        self.model = model
        self.vocab = vocab

    def __call__(
        self,
        text: Union[
            str,
            List[str],
            List[List[str]]
        ]
    ) -> Doc:
        """Convert input text to a spaCy Doc.

        text: The text to process. It can be presegmented or pretokenized:
            str             : raw text,
            List[str]       : presegmented text,
            List[List[str]] : pretokenized text.
        RETURNS: The spaCy Doc object.
        """
        if not text:
            return Doc(vocab=self.vocab)

        udpipe_sents = self.model(text=text) if text else [Sentence()]
        text = " ".join(s.getText() for s in udpipe_sents)
        tokens, heads = self._get_tokens_with_heads(udpipe_sents=udpipe_sents)

        words = []
        spaces = []
        pos = []
        tags = []
        morphs = []
        deps = []
        lemmas = []
        offset = 0
        is_aligned = self._check_aligned(text=text, tokens=tokens)
        if not is_aligned:
            text = ""
            for token in tokens:
                text += token.form
                if NO_SPACE not in token.misc:
                    text += " "
        for i, token in enumerate(tokens):
            span = text[offset:]
            if not span:
                break
            while len(span) and span[0].isspace():
                # If we encounter leading whitespace, skip one character ahead
                offset += 1
                span = text[offset:]
            words.append(token.form)
            # Make sure all strings are in the vocabulary
            pos.append(token.upostag or "")
            # CoNNL xpostag-s, custom for each UD treebank
            morphs.append(token.feats or "")
            tags.append(token.xpostag or "")
            deps.append(_spacy_dep(token.deprel) or "")
            lemmas.append(token.lemma or "")
            offset += len(token.form)
            span = text[offset:]
            if i == len(tokens) - 1 or NO_SPACE in token.misc:
                spaces.append(False)
            elif not is_aligned:
                spaces.append(True)
            else:
                next_token = tokens[i + 1]
                spaces.append(not span.startswith(next_token.form))
        doc = Doc(
            vocab=self.vocab,
            words=words,
            spaces=spaces,
            pos=pos,
            tags=tags,
            morphs=morphs,
            lemmas=lemmas,
            deps=deps,
            heads=[head + i for i, head in enumerate(heads)],
        )
        return doc

    def pipe(
        self,
        texts: Union[
            Iterable[str],
            Iterable[List[str]],
            Iterable[List[List[str]]]
        ]
    ) -> Iterable[Doc]:
        """Tokenize a stream of texts.

        texts: A sequence of unicode texts (raw, presegmented or pretokenized).
        n_process: Number of processes to use.
        YIELDS: A sequence of Doc objects, in order.
        """
        for text in texts:
            yield self(text)

    def _get_tokens_with_heads(
            self,
            udpipe_sents: List[Sentence]
    ) -> Tuple[List[str], List[int]]:
        """Flatten the tokens in the UDPipe sentence representations and extract
        the token indices of the sentence start tokens to is_sent_start set.

        udpipe_sents: The processed sentences.
        RETURNS: The tokens (words).
        """
        tokens = []
        heads = []
        offset = 0
        for sentence in udpipe_sents:
            words = sentence.words[1:]  # Ignore <root>
            for token in words:
                # Calculate the absolute token index in the doc,
                # then the *relative* index of the head, -1 for zero-indexed
                # and if the governor is 0 (root), we leave it at 0
                if token.head:
                    head = token.head + offset - len(tokens) - 1
                else:
                    head = 0
                heads.append(head)
                tokens.append(token)
            offset += len(words)
        return tokens, heads

    def _check_aligned(self, text: str, tokens: List[Word]) -> bool:
        """Check if tokens are aligned with text.

        text: Text to check.
        tokens: Tokens to check.
        RETURNS: True iff text and tokens are aligned.
        """
        token_texts = "".join(t.form for t in tokens)
        return re.sub(self._ws_pattern, "", text) == token_texts

    def to_disk(self, _path, **kwargs):
        return None

    def from_disk(self, _path, **kwargs):
        return self

    def to_bytes(self, **kwargs):
        return b""

    def from_bytes(self, _bytes_data, **kwargs):
        return self



import os

import setuptools


def get_version(fname: str) -> str:
    full_path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        "spacy_udpipe",
        fname
    )
    with open(full_path, "r") as fp:
        for l in fp:
            if l.startswith("__version__"):
                delim = '"' if '"' in l else "'"
                return l.split(delim)[1]
            else:
                raise RuntimeError(
                    "Unable to find version string."
                )


URL = "https://github.com/TakeLab/spacy-udpipe"

with open("README.md", "r") as f:
    readme = f.read()

setuptools.setup(
    name="spacy_udpipe",
    version=get_version("__init__.py"),
    description="Use fast UDPipe models directly in spaCy",
    long_description=readme,
    long_description_content_type="text/markdown",
    url=URL,
    author="TakeLab",
    author_email="takelab@fer.hr",
    license="MIT",
    keywords="nlp udpipe spacy python",
    packages=setuptools.find_packages(),
    install_requires=["spacy>=3.0.0,<4.0.0", "ufal.udpipe>=1.2.0"],
    python_requires=">=3.6",
    entry_points={
        "spacy_tokenizers": [
            "spacy_udpipe.PipelineAsTokenizer.v1 = spacy_udpipe:tokenizer.create_tokenizer",
        ]
    },
    tests_require=["pytest>=5.0.0"],
    package_data={"spacy_udpipe": ["./languages.json"], },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "SpaCy": "https://spacy.io/",
        "TakeLab": "http://takelab.fer.hr/",
        "UDPipe": "http://ufal.mff.cuni.cz/udpipe",
        "Source": URL,
        "Tracker": URL + "/issues",
    },
    zip_safe=False
)
